@@ -20,16 +20,17 @@ use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
     fcntl::OFlag,
-    sys::socket::{AddressFamily, MsgFlags, SockaddrLike, SockaddrStorage},
+    sys::socket::{AddressFamily, SockaddrLike, SockaddrStorage, UnixAddr},
     NixPath,
 };
 
 use crate::{
-    compat::{addr_family, PF_ALG, PF_INET, PF_INET6, PF_MAX, PF_NETLINK, PF_UNIX, PF_UNSPEC},
+    compat::{
+        addr_family, MsgFlags, PF_ALG, PF_INET, PF_INET6, PF_MAX, PF_NETLINK, PF_UNIX, PF_UNSPEC,
+    },
     config::HOOK_SCKCALLS,
     confine::{op2errno, op2name, scmp_arch_bits},
     fs::{fd_status_flags, file_type, safe_canonicalize, CanonicalPath, FileType, FsFlags},
-    hook::UNotifyEventRequest,
     kernel::net::{
         accept::handle_accept,
         bind::handle_bind,
@@ -38,11 +39,13 @@ use crate::{
         getsockname::handle_getsockname,
         getsockopt::handle_getsockopt,
         recvfrom::handle_recvfrom,
+        recvmsg::{handle_recvmmsg, handle_recvmmsg64, handle_recvmsg},
         sendmsg::{handle_sendmmsg, handle_sendmsg},
         sendto::handle_sendto,
-        socket::handle_socket,
+        socket::{handle_socket, handle_socketpair},
     },
     path::{XPath, XPathBuf},
+    req::UNotifyEventRequest,
     sandbox::{Action, Capability, SandboxGuard},
     warn,
 };
@@ -54,6 +57,7 @@ pub(crate) mod getpeername;
 pub(crate) mod getsockname;
 pub(crate) mod getsockopt;
 pub(crate) mod recvfrom;
+pub(crate) mod recvmsg;
 pub(crate) mod sendmsg;
 pub(crate) mod sendto;
 pub(crate) mod socket;
@@ -120,6 +124,11 @@ pub(crate) fn sys_socket(request: UNotifyEventRequest) -> ScmpNotifResp {
     syscall_network_handler(request, &req.data.args, 0x1)
 }
 
+pub(crate) fn sys_socketpair(request: UNotifyEventRequest) -> ScmpNotifResp {
+    let req = request.scmpreq;
+    syscall_network_handler(request, &req.data.args, 0x8)
+}
+
 pub(crate) fn sys_bind(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
     syscall_network_handler(request, &req.data.args, 0x2)
@@ -160,6 +169,23 @@ pub(crate) fn sys_recvfrom(request: UNotifyEventRequest) -> ScmpNotifResp {
     syscall_network_handler(request, &req.data.args, 0xc)
 }
 
+pub(crate) fn sys_recvmsg(request: UNotifyEventRequest) -> ScmpNotifResp {
+    let req = request.scmpreq;
+    syscall_network_handler(request, &req.data.args, 0x11)
+}
+
+pub(crate) fn sys_recvmmsg(request: UNotifyEventRequest) -> ScmpNotifResp {
+    let req = request.scmpreq;
+    syscall_network_handler(request, &req.data.args, 0x13)
+}
+
+pub(crate) fn sys_recvmmsg64(request: UNotifyEventRequest) -> ScmpNotifResp {
+    let req = request.scmpreq;
+    // recvmmsg_time64 does not have a socketcall(2) equivalent,
+    // therefore we use the dummy op `u8::MAX`.
+    syscall_network_handler(request, &req.data.args, u8::MAX)
+}
+
 pub(crate) fn sys_sendto(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
     syscall_network_handler(request, &req.data.args, 0xb)
@@ -180,28 +206,34 @@ pub(crate) fn sys_sendmmsg(request: UNotifyEventRequest) -> ScmpNotifResp {
 fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8) -> ScmpNotifResp {
     syscall_handler!(request, |request: UNotifyEventRequest| {
         let sandbox = request.get_sandbox();
-        let allow_safe_bind = sandbox.flags.allow_safe_bind();
-        let allow_safe_kcapi = sandbox.flags.allow_safe_kcapi();
-        let allow_unsupp_socket = sandbox.flags.allow_unsupp_socket();
-        let restrict_oob = !sandbox.flags.allow_unsafe_oob();
-        let restrict_mkbdev = !sandbox.flags.allow_unsafe_mkbdev();
+        let flags = *sandbox.flags;
+        let allow_safe_bind = flags.allow_safe_bind();
+        let allow_safe_kcapi = flags.allow_safe_kcapi();
+        let allow_unsupp_socket = flags.allow_unsupp_socket();
+        let restrict_oob = !flags.allow_unsafe_oob();
 
         let cap = match op {
             0x1 => {
                 // a. socket system call.
                 // b. socketcall -> socket indirection.
-                let flags = *sandbox.flags;
                 let nlfam = sandbox.netlink_families;
                 drop(sandbox); // drop read-lock before emulation.
-
                 return handle_socket(&request, args, flags, nlfam);
             }
-            0x5 | 0x6 | 0x7 | 0xc | 0x12 => {
+            0x8 => {
+                // a. socketpair system call.
+                // b. socketcall -> socketpair indirection.
+                // As of version 3.42.0, this is checked for bind
+                // access with the dummy `!unnamed' path.
+                return handle_socketpair(&request, sandbox, args, 0x8);
+            }
+            0x5 | 0x6 | 0x7 | 0xc | 0x11 | 0x12 | 0x13 | u8::MAX => {
                 // accept, accept4, getsockname, getpeername:
                 //
                 // accept{,4} are IP blocklist only.
                 // get{peer,sock}name are informational.
                 // recvfrom is informational.
+                // recvmsg, recvmmsg, and recvmmsg_time64 are informational.
                 Capability::empty()
             }
             0x2 /* bind */ => Capability::CAP_NET_BIND,
@@ -251,31 +283,29 @@ fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8
                 // recvfrom is used for informational purposes only.
                 return handle_recvfrom(fd, args, &request, restrict_oob);
             }
+            0x11 => {
+                // recvmsg is used for informational purposes only.
+                return handle_recvmsg(fd, &request, args, flags);
+            }
+            0x13 => {
+                // recvmsg is used for informational purposes only.
+                return handle_recvmmsg(fd, &request, args, flags);
+            }
+            u8::MAX => {
+                // recvmmsg_time64 is used for informational purposes only.
+                return handle_recvmmsg64(fd, &request, args, flags);
+            }
             0xf => {
                 // getsockopt is used for informational purposes only.
                 return handle_getsockopt(fd, &request, args);
             }
             0x10 => {
                 // sendmsg uses a different data structure, so we handle it in its own branch.
-                return handle_sendmsg(
-                    fd,
-                    &request,
-                    args,
-                    allow_unsupp_socket,
-                    restrict_oob,
-                    restrict_mkbdev,
-                );
+                return handle_sendmsg(fd, &request, args, flags);
             }
             0x14 => {
                 // sendmmsg uses a different data structure, so we handle it in its own branch.
-                return handle_sendmmsg(
-                    fd,
-                    &request,
-                    args,
-                    allow_unsupp_socket,
-                    restrict_oob,
-                    restrict_mkbdev,
-                );
+                return handle_sendmmsg(fd, &request, args, flags);
             }
             _ => {} // fall through.
         }
@@ -297,7 +327,16 @@ fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8
         } // else we have a valid address to check for access.
 
         let sandbox = request.get_sandbox();
-        let argaddr = get_addr(&request, addr_remote, addr_len)?;
+        let argaddr = if op == 0x2 && addr_len as usize == size_of::<libc::sa_family_t>() {
+            // unix(7): "If a bind(2) call specifies addrlen as `sizeof(sa_family_t)`,
+            // or [...], then the socket is autobound to an abstract address".
+            let addr = UnixAddr::new_unnamed();
+            // SAFETY: `addr` is constructed using new_unnamed.
+            #[expect(clippy::disallowed_methods)]
+            unsafe { SockaddrStorage::from_raw(addr.as_ptr().cast(), Some(addr.len())) }.unwrap()
+        } else {
+            get_addr(&request, addr_remote, addr_len)?
+        };
         let (addr, root) = canon_addr(&request, &sandbox, &argaddr, cap)?;
         match addr_family(&addr) {
             PF_UNIX | PF_INET | PF_INET6 => {
@@ -747,6 +786,59 @@ pub(crate) fn sandbox_addr_notsup(sandbox: &SandboxGuard) -> Result<(), Errno> {
     }
 }
 
+/// Process an unnamed socket with the given socket family.
+///
+/// Used at socketpair(2) boundary for AF_UNIX sockets.
+#[expect(clippy::cognitive_complexity)]
+pub(crate) fn sandbox_addr_unnamed(
+    request: &UNotifyEventRequest,
+    sandbox: &SandboxGuard,
+    op: u8,
+) -> Result<(), Errno> {
+    // socketpair
+    assert_eq!(
+        op, 0x8,
+        "BUG: sandbox_addr_unnamed called with invalid op:{op:#x}, report a bug!"
+    );
+
+    // Check for access.
+    let caps = Capability::CAP_NET_BIND;
+    let name = XPath::from_bytes(b"!unnamed");
+
+    if sandbox.getcaps(caps).is_empty() {
+        // Sandboxing is off.
+        return Ok(());
+    }
+    let (action, filter) = sandbox.check_unix(caps, name);
+
+    if !filter {
+        let sys = op2name(op);
+        if sandbox.verbose {
+            warn!("ctx": "access", "cap": caps, "act": action,
+                "sys": sys, "unix": &name, "abs": true,
+                "tip": format!("configure `allow/{caps}+{name}'"),
+                "req": request);
+        } else {
+            warn!("ctx": "access", "cap": caps, "act": action,
+                "sys": sys, "unix": &name, "abs": true,
+                "tip": format!("configure `allow/{caps}+{name}'"),
+                "pid": request.scmpreq.pid);
+        }
+    }
+
+    match action {
+        Action::Allow | Action::Warn => Ok(()),
+        Action::Deny | Action::Filter => Err(op2errno(op)),
+        Action::Panic => panic!(),
+        Action::Exit => std::process::exit(op2errno(op) as i32),
+        action => {
+            // Stop|Kill
+            let _ = request.kill(action);
+            Err(op2errno(op))
+        }
+    }
+}
+
 pub(crate) fn to_msgflags(arg: u64) -> Result<MsgFlags, Errno> {
     let flags = arg.try_into().or(Err(Errno::EINVAL))?;
     MsgFlags::from_bits(flags).ok_or(Errno::EINVAL)
