@@ -150,6 +150,8 @@ pub mod landlock_policy;
 pub mod log;
 /// magic symlink utilities
 pub mod magic;
+/// Interface to Open File Description locks
+pub mod ofd;
 /// /proc and syd(2) nom parsers
 pub mod parsers;
 /// Path handling for UNIX
@@ -158,6 +160,10 @@ pub mod path;
 pub mod proc;
 /// ptrace(2) utilities
 pub mod ptrace;
+/// seccomp(2) notify request handling
+pub mod req;
+/// Utilities to handle restarting syscalls
+pub mod retry;
 /// OS Random Number Generator (RNG) interface
 pub mod rng;
 /// Sandbox configuration
@@ -191,11 +197,6 @@ pub mod wordexp;
 #[expect(clippy::arithmetic_side_effects)]
 #[expect(clippy::undocumented_unsafe_blocks)]
 pub mod caps;
-// Used by Landlock crate.
-// lazy_static is a dev-only dependency.
-#[cfg(test)]
-#[macro_use]
-extern crate lazy_static;
 /// Interface to LandLock LSM
 #[expect(missing_docs)]
 #[expect(unused_imports)]
@@ -1169,7 +1170,7 @@ pub fn set_sigpipe_dfl() -> Result<(), Errno> {
 #[cfg(all(not(feature = "prof"), target_pointer_width = "64"))]
 #[allow(non_upper_case_globals)]
 pub fn config_mimalloc() {
-    // Extended feture is not usable (breaks on loongarch64),
+    // Extended feature is not usable (breaks on loongarch64),
     // therefore we define a subset of mimalloc API here.
     #[allow(non_camel_case_types)]
     type mi_option_t = libc::c_int;
