@@ -52,10 +52,13 @@ use nix::{
         uio::{process_vm_readv, process_vm_writev, RemoteIoVec},
     },
     unistd::{getpid, lseek64, write, Gid, Pid, Uid, Whence},
+    NixPath,
 };
 use procfs_core::{
     net::UnixNetEntries,
-    process::{LimitValue, MMPermissions, MMapPath, MemoryMaps, Namespace, Namespaces},
+    process::{
+        LimitValue, MMPermissions, MMapPath, MemoryMaps, Namespace, Namespaces, SmapsRollup,
+    },
     CryptoTable, FromBufRead,
 };
 
@@ -66,14 +69,18 @@ use crate::{
     cookie::{PROCMAP_QUERY_COOKIE_ARG3, PROCMAP_QUERY_COOKIE_ARG4, PROCMAP_QUERY_COOKIE_ARG5},
     elf::ElfType,
     err::{err2no, proc_error_to_errno},
-    fs::{is_open_fd, parse_fd, read_all, retry_on_eintr, safe_open, write_all, AT_BADFD},
+    fs::{is_open_fd, parse_fd, read_all, safe_open, write_all, AT_BADFD},
     hash::{SydHashMap, SydHashSet, SydIndexSet},
     parsers::{
         map_result,
-        proc::{parse_max_open_files, parse_stat, parse_statm, parse_status},
+        proc::{
+            parse_max_open_files, parse_stat, parse_statm, parse_status, parse_status_tgid,
+            parse_status_umask,
+        },
         read_to_end,
     },
     path::{XPathBuf, PATH_MAX},
+    retry::retry_on_eintr,
     rng::randint,
     sigset::SydSigSet,
     XPath,
@@ -143,8 +150,6 @@ impl From<MMPermissions> for ProcmapQueryFlags {
     }
 }
 
-#[repr(C)]
-#[derive(Debug, Clone, Copy)]
 /// Input/output argument structure passed into the _ioctl_(2) call. It can
 /// be used to query a set of VMAs (Virtual Memory Areas) of a process.
 ///
@@ -166,6 +171,8 @@ impl From<MMPermissions> for ProcmapQueryFlags {
 /// _proc_pid_maps_(5), but `procmap_query` provides more querying flexibility.
 /// There are no consistency guarantees between subsequent _ioctl_(2) calls,
 /// but data returned for a matched VMA is self-consistent.
+#[repr(C)]
+#[derive(Copy, Clone, Debug, PartialEq, Eq, Hash)]
 pub struct ProcmapQuery {
     /// Query struct size, for backwards/forward compatibility.
     pub size: u64,
@@ -320,7 +327,7 @@ pub fn procmap_query<Fd: AsFd>(
 /// All values are in units of pages.
 ///
 /// See `man 5 proc` and `Linux/fs/proc/array.c`.
-#[derive(Debug, Default, PartialEq, Eq, Hash)]
+#[derive(Copy, Clone, Debug, Default, PartialEq, Eq, Hash)]
 pub struct Statm {
     /// Total virtual memory size.
     pub size: u64,
@@ -329,10 +336,12 @@ pub struct Statm {
 /// Process status information.
 ///
 /// See `man 5 proc` and `Linux/fs/proc/array.c`.
-#[derive(Debug, Default, PartialEq, Eq, Hash)]
+#[derive(Copy, Clone, Debug, Default, PartialEq, Eq, Hash)]
 pub struct Stat {
     /// Number of threads in this process (since Linux 2.6).
     pub num_threads: u64,
+    /// The address above which program heap can be expanded with brk(2).
+    pub startbrk: u64,
     /// The address of the start (i.e., bottom) of the stack.
     pub startstack: u64,
     /// The controlling terminal of the process. (The minor device
@@ -344,14 +353,12 @@ pub struct Stat {
 /// Process status information.
 ///
 /// See `man 5 proc` and `Linux/fs/proc/array.c`.
-#[derive(Default, Debug, PartialEq, Eq, Hash)]
+#[derive(Copy, Clone, Debug, PartialEq, Eq, Hash)]
 pub struct Status {
-    /// Filename of the executable.
-    pub command: XPathBuf,
     /// File mode creation mask (since Linux 4.7).
-    pub umask: libc::mode_t,
+    pub umask: Mode,
     /// Process ID (i.e., Thread Group ID).
-    pub pid: libc::pid_t,
+    pub pid: Pid,
     /// Number of signals pending for the thread (see pthreads(7)).
     pub sig_pending_thread: SydSigSet,
     /// Number of signals pending for the process (see signal(7)).
@@ -364,6 +371,22 @@ pub struct Status {
     pub sig_caught: SydSigSet,
 }
 
+// Default is used by parsers.
+// Pid has no default therefore this is needed.
+impl Default for Status {
+    fn default() -> Self {
+        Self {
+            umask: Mode::empty(),
+            pid: Pid::from_raw(0),
+            sig_pending_thread: SydSigSet::default(),
+            sig_pending_process: SydSigSet::default(),
+            sig_blocked: SydSigSet::default(),
+            sig_ignored: SydSigSet::default(),
+            sig_caught: SydSigSet::default(),
+        }
+    }
+}
+
 /// Safe `OFlags` to use when opening proc(5) files.
 pub const OFLAG_PROC: OFlag =
     OFlag::from_bits_retain(libc::O_RDONLY | libc::O_CLOEXEC | libc::O_NOFOLLOW | libc::O_NOCTTY);
@@ -385,7 +408,7 @@ fn proc_errno(err: Errno) -> Errno {
 }
 
 /// Function to safely open a proc(5) file and return an `OwnedFd`.
-pub fn safe_open_proc(path: &XPath) -> Result<OwnedFd, Errno> {
+pub fn safe_open_proc<P: NixPath + ?Sized>(path: &P) -> Result<OwnedFd, Errno> {
     safe_open(PROC_FILE(), path, OFLAG_PROC, RFLAG_PROC)
 }
 
@@ -501,14 +524,28 @@ pub fn proc_status_read<Fd: AsFd>(fd: Fd) -> Result<Status, Errno> {
 
 /// Retrieves the thread group ID (TGID) for the specified thread ID (TID).
 pub fn proc_tgid(tid: Pid) -> Result<Pid, Errno> {
-    proc_status(tid).map(|stat| stat.pid).map(Pid::from_raw)
+    // This function is a hot-path, so we provide a specialized impl.
+    // proc_status(tid).map(|stat| stat.pid)
+    let fd = proc_status_open(tid)?;
+
+    // Read up to 192 bytes.
+    let mut data = [0u8; 192];
+    let nread = read_all(fd, &mut data)?;
+
+    map_result(parse_status_tgid(&data[..nread]))
 }
 
 /// Get the umask of the given `Pid`.
 pub fn proc_umask(pid: Pid) -> Result<Mode, Errno> {
-    proc_status(pid)
-        .map(|stat| stat.umask)
-        .map(Mode::from_bits_truncate)
+    // This function is a hot-path, so we provide a specialized impl.
+    // proc_status(pid).map(|stat| stat.umask)
+    let fd = proc_status_open(pid)?;
+
+    // Read up to 160 bytes.
+    let mut data = [0u8; 160];
+    let nread = read_all(fd, &mut data)?;
+
+    map_result(parse_status_umask(&data[..nread]))
 }
 
 /// Locate stack and return the stack memory address range.
@@ -734,6 +771,45 @@ pub fn proc_smaps(pid: Pid) -> Result<Vec<SydMemoryMap>, Errno> {
         .map_err(proc_errno)
 }
 
+/// Return the rollup memory map of the given process.
+///
+/// This is more efficient than using `proc_smaps` and iterating over maps.
+/// See: https://www.kernel.org/doc/Documentation/ABI/testing/procfs-smaps_rollup
+pub fn proc_smaps_rollup(pid: Pid) -> Result<SydMemoryMap, Errno> {
+    let mut path = XPathBuf::from_pid(pid);
+    path.push(b"smaps_rollup");
+
+    // Safely open proc(5) file.
+    let reader = safe_open_proc(&path)
+        .map(File::from)
+        .map(BufReader::new)
+        .map_err(proc_errno)?;
+
+    SmapsRollup::from_buf_read(reader)
+        .map_err(|err| proc_error_to_errno(&err).unwrap_or(Errno::ENOSYS))
+        .map_err(proc_errno)
+        .and_then(|maps| {
+            maps.memory_map_rollup
+                .into_iter()
+                .next()
+                .map(SydMemoryMap)
+                .ok_or(Errno::ENOMEM)
+        })
+}
+
+/// Calculate total memory usage of a process using _proc_pid_smaps_rollup(5).
+///
+/// See: https://www.kernel.org/doc/Documentation/ABI/testing/procfs-smaps_rollup
+pub fn proc_mem(pid: Pid) -> Result<u64, Errno> {
+    let map = proc_smaps_rollup(pid)?;
+    let mut sum = 0u64;
+    for key in ["Pss", "Private_Dirty", "Shared_Dirty"] {
+        let val = map.0.extension.map.get(key).copied().unwrap_or(0);
+        sum = sum.saturating_add(val);
+    }
+    Ok(sum)
+}
+
 /// Returns the memory map of an address by reading _proc_pid_maps_(5).
 ///
 /// Returns EFAULT if the address does not fall within any range.
@@ -758,7 +834,7 @@ pub fn proc_pipemax() -> Result<libc::c_int, Errno> {
 
 /// Open a FD to /proc/sys/fs/pipe-max-size and return an OwnedFd.
 pub fn proc_pipemax_open() -> Result<OwnedFd, Errno> {
-    safe_open_proc(XPath::from_bytes(b"sys/fs/pipe-max-size"))
+    safe_open_proc("sys/fs/pipe-max-size")
 }
 
 /// Retrieve the system pipe max limit from the given Fd.
@@ -775,12 +851,7 @@ pub fn proc_fs_file_max() -> Result<u64, Errno> {
     // SAFETY: Informational command,
     // runs before syd::config::proc_init.
     #[expect(clippy::disallowed_methods)]
-    let fd = nix::fcntl::openat(
-        AT_BADFD,
-        XPath::from_bytes(b"/proc/sys/fs/file-max"),
-        OFLAG_PROC,
-        Mode::empty(),
-    )?;
+    let fd = nix::fcntl::openat(AT_BADFD, "/proc/sys/fs/file-max", OFLAG_PROC, Mode::empty())?;
 
     // Read up to 24 bytes + 1 bytes for \n.
     let mut data = [0u8; 25];
@@ -794,12 +865,7 @@ pub fn proc_fs_nr_open() -> Result<u64, Errno> {
     // SAFETY: Informational command,
     // runs before syd::config::proc_init.
     #[expect(clippy::disallowed_methods)]
-    let fd = nix::fcntl::openat(
-        AT_BADFD,
-        XPath::from_bytes(b"/proc/sys/fs/nr_open"),
-        OFLAG_PROC,
-        Mode::empty(),
-    )?;
+    let fd = nix::fcntl::openat(AT_BADFD, "/proc/sys/fs/nr_open", OFLAG_PROC, Mode::empty())?;
 
     // Read up to 24 bytes + 1 bytes for \n.
     let mut data = [0u8; 25];
@@ -815,7 +881,7 @@ pub fn proc_mmap_min_addr() -> Result<u64, Errno> {
     #[expect(clippy::disallowed_methods)]
     let fd = nix::fcntl::openat(
         AT_BADFD,
-        XPath::from_bytes(b"/proc/sys/vm/mmap_min_addr"),
+        "/proc/sys/vm/mmap_min_addr",
         OFLAG_PROC,
         Mode::empty(),
     )?;
@@ -834,7 +900,7 @@ pub fn proc_pid_max(max: u64) -> Result<(), Errno> {
     #[expect(clippy::disallowed_methods)]
     let fd = nix::fcntl::openat(
         AT_BADFD,
-        XPath::from_bytes(b"/proc/sys/kernel/pid_max"),
+        "/proc/sys/kernel/pid_max",
         OFLAG_PROC | OFlag::O_WRONLY,
         Mode::empty(),
     )?;
@@ -850,7 +916,7 @@ pub fn proc_kernel_randomize_va_space() -> Result<u8, Errno> {
     #[expect(clippy::disallowed_methods)]
     let fd = nix::fcntl::openat(
         AT_BADFD,
-        XPath::from_bytes(b"/proc/sys/kernel/randomize_va_space"),
+        "/proc/sys/kernel/randomize_va_space",
         OFLAG_PROC,
         Mode::empty(),
     )?;
@@ -1042,67 +1108,17 @@ pub fn proc_task_limit(pid: Pid, max: u64) -> Result<bool, Errno> {
     Ok(false)
 }
 
-/// Checks whether process memory usage is within the give maximum.
-///
-/// This function uses the `procfs` crate to obtain detailed memory maps
-/// from `/proc/[pid]/smaps`. It sums multiple memory usage values reported in these maps
-/// to calculate a more comprehensive total memory usage.
-///
-/// # Returns
-///
-/// This function returns a `Result<bool, Errno>`. It returns Ok(true)
-/// if the limit was exceeded Ok(false) otherwise. On failure, it
-/// returns `Errno`.
-///
-/// # Errors
-///
-/// This function returns an error if it fails to retrieve the process's memory maps,
-/// typically due to insufficient permissions or an invalid process ID.
-pub fn proc_mem_limit(pid: Pid, max: u64) -> Result<bool, Errno> {
-    match proc_smaps(pid) {
-        Ok(maps) => {
-            let mut total_size: u64 = 0;
-            for map in &maps {
-                match &map.0.pathname {
-                    MMapPath::Path(_)
-                    | MMapPath::Anonymous
-                    | MMapPath::Stack
-                    | MMapPath::Other(_) => {
-                        let pss = map.0.extension.map.get("Pss").copied().unwrap_or(0);
-                        let private_dirty = map
-                            .0
-                            .extension
-                            .map
-                            .get("Private_Dirty")
-                            .copied()
-                            .unwrap_or(0);
-                        let shared_dirty = map
-                            .0
-                            .extension
-                            .map
-                            .get("Shared_Dirty")
-                            .copied()
-                            .unwrap_or(0);
-
-                        total_size = total_size.saturating_add(
-                            pss.saturating_add(private_dirty)
-                                .saturating_add(shared_dirty),
-                        );
-
-                        // Stop processing if total size exceeds or equals max
-                        if total_size >= max {
-                            return Ok(true);
-                        }
-                    }
-                    _ => (),
-                }
-            }
-
-            // If we're at this point, we did not hit the limit.
-            Ok(false)
-        }
-        Err(_) => Err(Errno::last()),
-    }
+/// Represents an executable file record from proc_pid_maps(5).
+#[derive(Clone, Debug, Eq, PartialEq, Ord, PartialOrd, Hash)]
+pub struct SydExecMap {
+    /// Path to the executable
+    pub path: XPathBuf,
+    /// Inode of the executable
+    pub inode: u64,
+    /// Major device ID of the executable
+    pub dev_major: i32,
+    /// Minor device ID of the executable
+    pub dev_minor: i32,
 }
 
 /// Collects all unique paths with executable permissions from the
@@ -1110,12 +1126,11 @@ pub fn proc_mem_limit(pid: Pid, max: u64) -> Result<bool, Errno> {
 ///
 /// # Returns
 ///
-/// A `Result` containing a vector of `(XPathBuf,u64,i32,i32)` objects
+/// A `Result` containing a vector of `SydExecMap` objects
 /// for all unique paths with executable permissions along with their
 /// device ID and inode. On error, it returns an `Errno`. If no executable
 /// is found in maps, returns `Errno::ENOEXEC`.
-#[expect(clippy::type_complexity)]
-pub fn proc_executables(pid: Pid) -> Result<Vec<(XPathBuf, u64, i32, i32)>, Errno> {
+pub fn proc_executables(pid: Pid) -> Result<Vec<SydExecMap>, Errno> {
     if *crate::config::HAVE_PROCMAP_QUERY {
         // Use the efficient PROCMAP_QUERY ioctl(2) on Linux>=6.11.
         return procmap_query_executables(pid);
@@ -1125,7 +1140,7 @@ pub fn proc_executables(pid: Pid) -> Result<Vec<(XPathBuf, u64, i32, i32)>, Errn
 
     // Create a SydIndexSet to store unique executable paths while
     // preserving insertion order.
-    let mut paths = SydIndexSet::default();
+    let mut set = SydIndexSet::default();
 
     // Iterate over each memory map entry.
     for map in maps {
@@ -1134,18 +1149,24 @@ pub fn proc_executables(pid: Pid) -> Result<Vec<(XPathBuf, u64, i32, i32)>, Errn
             // Check if the map has executable permissions.
             if map.0.perms.contains(MMPermissions::EXECUTE) {
                 // Insert the path into the SydIndexSet.
-                paths.insert((path.into(), map.0.inode, map.0.dev.0, map.0.dev.1));
+                let exe = SydExecMap {
+                    path: path.into(),
+                    inode: map.0.inode,
+                    dev_major: map.0.dev.0,
+                    dev_minor: map.0.dev.1,
+                };
+                set.insert(exe);
             }
         }
     }
 
     // Check for no match.
-    if paths.is_empty() {
+    if set.is_empty() {
         return Err(Errno::ENOEXEC);
     }
 
     // Convert the SydIndexSet to a vector and return.
-    Ok(paths.into_iter().collect())
+    Ok(set.into_iter().collect())
 }
 
 /// Collects all unique paths with executable permissions from the
@@ -1153,12 +1174,11 @@ pub fn proc_executables(pid: Pid) -> Result<Vec<(XPathBuf, u64, i32, i32)>, Errn
 ///
 /// # Returns
 ///
-/// A `Result` containing a vector of `(XPathBuf, u64, i32, i32)` objects
+/// A `Result` containing a vector of `SydExecMap` objects
 /// for all unique paths with executable permissions along with their
 /// inode and device numbers. On error, returns an `Errno`. If no executable
 /// mappings are found, returns `Errno::ENOEXEC`.
-#[expect(clippy::type_complexity)]
-pub fn procmap_query_executables(pid: Pid) -> Result<Vec<(XPathBuf, u64, i32, i32)>, Errno> {
+pub fn procmap_query_executables(pid: Pid) -> Result<Vec<SydExecMap>, Errno> {
     // Open /proc/$pid/maps.
     let maps = proc_maps_open(pid)?;
 
@@ -1169,7 +1189,7 @@ pub fn procmap_query_executables(pid: Pid) -> Result<Vec<(XPathBuf, u64, i32, i3
 
     let mut path = [0u8; PATH_MAX];
     let mut addr = 0u64;
-    let mut paths = SydIndexSet::default();
+    let mut set = SydIndexSet::default();
 
     loop {
         match procmap_query(&maps, flags, addr, Some(&mut path), None) {
@@ -1186,7 +1206,13 @@ pub fn procmap_query_executables(pid: Pid) -> Result<Vec<(XPathBuf, u64, i32, i3
                     .map(XPathBuf::from)
                     .map_err(|_| Errno::EINVAL)?;
                 #[expect(clippy::cast_possible_wrap)]
-                paths.insert((path, q.inode, q.dev_major as i32, q.dev_minor as i32));
+                let exe = SydExecMap {
+                    path,
+                    inode: q.inode,
+                    dev_major: q.dev_major as i32,
+                    dev_minor: q.dev_minor as i32,
+                };
+                set.insert(exe);
 
                 addr = q.vma_end;
             }
@@ -1196,12 +1222,12 @@ pub fn procmap_query_executables(pid: Pid) -> Result<Vec<(XPathBuf, u64, i32, i3
     }
 
     // Check for no match.
-    if paths.is_empty() {
+    if set.is_empty() {
         return Err(Errno::ENOEXEC);
     }
 
     // Convert the SydIndexSet to a vector and return.
-    Ok(paths.into_iter().collect())
+    Ok(set.into_iter().collect())
 }
 
 /// This data structure represents a Virtual Memory Area.
@@ -1358,6 +1384,8 @@ pub fn proc_namespaces(pid: Pid) -> Result<Namespaces, Errno> {
     // Read directory entries using getdents64 in a loop.
     #[expect(clippy::disallowed_types)]
     let mut namespaces = std::collections::HashMap::new();
+    let mut seen_dot = false;
+    let mut seen_dotdot = false;
     loop {
         let mut entries = match getdents64(&dirfd, DIRENT_BUF_SIZE) {
             Ok(iter) => iter,
@@ -1366,7 +1394,12 @@ pub fn proc_namespaces(pid: Pid) -> Result<Namespaces, Errno> {
         };
 
         for entry in &mut entries {
-            if entry.is_dot() {
+            if !seen_dot && entry.is_dot() {
+                seen_dot = true;
+                continue;
+            }
+            if !seen_dotdot && entry.is_dotdot() {
+                seen_dotdot = true;
                 continue;
             }
 
@@ -1405,7 +1438,7 @@ pub fn proc_crypto() -> Result<CryptoTable, Errno> {
 
 /// Open a FD to _proc_crypto_(5) and return an OwnedFd.
 pub fn proc_crypto_open() -> Result<OwnedFd, Errno> {
-    safe_open_proc(XPath::from_bytes(b"crypto"))
+    safe_open_proc("crypto")
 }
 
 /// Parses /proc/crypto from given Fd and returns the table of crypto algorithms.
@@ -1413,10 +1446,15 @@ pub fn proc_crypto_read(fd: OwnedFd) -> Result<CryptoTable, Errno> {
     CryptoTable::from_buf_read(BufReader::new(File::from(fd))).or(Err(Errno::EPERM))
 }
 
-/// Parses /proc/net/unix and returns the inodes of all UNIX domain sockets.
-pub fn proc_unix_get_inodes() -> Result<SydHashSet<u64>, Errno> {
-    let path = XPath::from_bytes(b"net/unix");
-    let file = safe_open_proc(path).map(File::from).map(BufReader::new)?;
+/// Parses /proc/pid/net/unix, return inodes.
+///
+/// Returns the inodes of all UNIX domain sockets.
+pub fn proc_unix_inodes(pid: Pid) -> Result<SydHashSet<u64>, Errno> {
+    let mut path = XPathBuf::from_pid(pid);
+    path.push(b"net");
+    path.push(b"unix");
+
+    let file = safe_open_proc(&path).map(File::from).map(BufReader::new)?;
     Ok(UnixNetEntries::from_buf_read(file)
         .map(|e| e.0)
         .or(Err(Errno::EPERM))?
@@ -1960,7 +1998,7 @@ pub fn proc_kernel_taint() -> Result<KernelTaintFlags, Errno> {
     #[expect(clippy::disallowed_methods)]
     let fd = nix::fcntl::openat(
         AT_BADFD,
-        XPath::from_bytes(b"/proc/sys/kernel/tainted"),
+        "/proc/sys/kernel/tainted",
         OFLAG_PROC,
         Mode::empty(),
     )?;
@@ -1999,6 +2037,9 @@ mod tests {
             return;
         }
 
+        let result = proc_tgid(Pid::from_raw(i32::MAX));
+        assert!(result.is_err(), "{result:?}");
+
         let result = proc_umask(Pid::from_raw(i32::MAX));
         assert!(result.is_err(), "{result:?}");
     }
@@ -2372,15 +2413,16 @@ mod tests {
         }
 
         // Ensure uniqueness and preserved order.
-        let list = proc_executables(Pid::this()).unwrap();
+        let bins = proc_executables(Pid::this()).unwrap();
         let mut seen = SydIndexSet::default();
-        for (path, _, _, _) in &list {
+        for bin in &bins {
+            let path = &bin.path;
             assert!(seen.insert(path.clone()), "duplicate path {path}!");
         }
 
         // Check that insertion order matches returned order.
         let collected: Vec<_> = seen.into_iter().collect();
-        let returned: Vec<_> = list.iter().map(|(p, _, _, _)| p.clone()).collect();
+        let returned: Vec<_> = bins.iter().map(|bin| bin.path.clone()).collect();
         assert_eq!(collected, returned);
     }
 
@@ -2391,8 +2433,9 @@ mod tests {
         }
 
         // Returned paths should exist on filesystem and have exec bit.
-        let list = proc_executables(Pid::this()).unwrap();
-        for (path, _, _, _) in list {
+        let bins = proc_executables(Pid::this()).unwrap();
+        for bin in bins {
+            let path = bin.path;
             let md = std::fs::metadata(&path).expect("path does not exist");
             let perms = md.permissions().mode();
             assert!(
