@@ -70,13 +70,12 @@ use data_encoding::HEXLOWER;
 use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
-    fcntl::{open, OFlag},
+    fcntl::OFlag,
     libc::setdomainname,
-    pty::{grantpt, posix_openpt, ptsname_r, unlockpt},
+    pty::{grantpt, posix_openpt, unlockpt},
     sched::{unshare, CloneFlags},
     sys::{
         socket::{bind, socket, AddressFamily, SockFlag, SockType, SockaddrIn, SockaddrIn6},
-        stat::Mode,
         wait::{Id, WaitPidFlag},
     },
     time::{clock_gettime, ClockId},
@@ -87,10 +86,10 @@ use syd::{
     caps,
     compat::{waitid, WaitStatus},
     config::*,
-    confine::ExportMode,
+    confine::{confine_scmp_madvise, ExportMode},
     err::err2no,
     error,
-    fs::{closeexcept, format_clone_flags, format_clone_names, pidfd_open, set_cloexec},
+    fs::{closeexcept, format_clone_flags, format_clone_names, openpts, pidfd_open, set_cloexec},
     hash::{get_at_random_hex, hash, HashAlgorithm, SydHashMap, SydHashSet},
     hook::Supervisor,
     ignore_signals, info,
@@ -585,15 +584,13 @@ syd::main! {
 
         // Step 2: Open a pseudoterminal device without O_CLOEXEC.
         let pty_m = posix_openpt(OFlag::O_RDWR | OFlag::O_NOCTTY | OFlag::O_NOFOLLOW)?;
+
+        // Grant access and unlock.
         grantpt(&pty_m)?;
         unlockpt(&pty_m)?;
-        let pty_s = ptsname_r(&pty_m).map(XPathBuf::from)?;
-        #[expect(clippy::disallowed_methods)]
-        let pty_s = open(
-            &pty_s,
-            OFlag::O_RDWR | OFlag::O_NOCTTY | OFlag::O_NOFOLLOW | OFlag::O_CLOEXEC,
-            Mode::empty(),
-        )?;
+
+        // Open peer device.
+        let pty_s = openpts(&pty_m, OFlag::O_RDWR | OFlag::O_NOCTTY | OFlag::O_CLOEXEC)?;
 
         // Step 3: Prepare environment of the syd-pty process.
         // Filter the environment variables to only include the list below:
@@ -1085,7 +1082,7 @@ syd::main! {
         "exit_group",
         "waitid",
         "brk",
-        "madvise",
+        //"madvise", advice are confined.
         "mremap",
         "munmap",
         "sigaction",
@@ -1112,6 +1109,9 @@ syd::main! {
         }
     }
 
+    // Allow safe madvise(2) advice.
+    confine_scmp_madvise(&mut ctx)?;
+
     // Allow write(2) to standard error.
     if let Ok(syscall) = ScmpSyscall::from_name("write") {
         ctx.add_rule_conditional(
