@@ -35,8 +35,8 @@ use crate::cookie::FTRUNCATE64_COOKIE_ARG3;
 use crate::{
     config::*,
     confine::{
-        confine_scmp_clone, confine_scmp_clone3, confine_scmp_fadvise, confine_scmp_write,
-        scmp_add_setid_rules, ExportMode,
+        confine_scmp_clone, confine_scmp_clone3, confine_scmp_fadvise, confine_scmp_madvise,
+        confine_scmp_write, scmp_add_setid_rules, ExportMode,
     },
     cookie::{
         safe_ftruncate64, safe_pipe2, ACCEPT4_COOKIE_ARG4, ACCEPT4_COOKIE_ARG5,
@@ -47,13 +47,15 @@ use crate::{
     debug,
     err::{err2no, SydJoinHandle, SydResult},
     error,
-    fs::{lock_fd, retry_on_eintr, retry_on_intr, seal_memfd, FileInfo},
+    fs::{seal_memfd, FileInfo},
     hash::{
         aes_ctr_enc, aes_ctr_init, hmac_sha256_feed, hmac_sha256_fini, hmac_sha256_init,
         SydHashMap, BLOCK_SIZE, HMAC_TAG_SIZE, IV, IV_SIZE, SYD3_HDR_SIZE,
     },
     info,
+    ofd::lock_fd,
     path::{XPath, XPathBuf},
+    retry::{retry_on_eintr, retry_on_intr},
     sandbox::Flags,
 };
 
@@ -169,8 +171,8 @@ impl AesWorker {
         ctx.set_ctl_nnp(true)?;
 
         // Disable Speculative Store Bypass mitigations
-        // with trace/allow_unsafe_spec_exec:1
-        ctx.set_ctl_ssb(flags.allow_unsafe_spec_exec())?;
+        // with trace/allow_unsafe_exec_speculative:1
+        ctx.set_ctl_ssb(flags.allow_unsafe_exec_speculative())?;
 
         // DO NOT synchronize filter to all threads.
         // Main thread will confine itself.
@@ -248,6 +250,9 @@ impl AesWorker {
         // Allow fadvise family syscalls.
         confine_scmp_fadvise(&mut ctx)?;
 
+        // Allow safe madvise(2) advice.
+        confine_scmp_madvise(&mut ctx)?;
+
         // ftruncate{,64}(2) may be used only with syscall argument cookies.
         let sysname = "ftruncate";
         #[expect(clippy::useless_conversion)]
@@ -341,6 +346,7 @@ impl AesWorker {
         // Restrict pipe2(2) using syscall argument cookies.
         let sysname = "pipe2";
         #[expect(clippy::cast_sign_loss)]
+        #[expect(clippy::useless_conversion)]
         if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
             let mut rules = vec![scmp_cmp!($arg1 == OFlag::O_CLOEXEC.bits() as u64)];
             if !flags.allow_unsafe_nocookie() {
@@ -867,7 +873,7 @@ impl AesWorker {
         let last_block_offset = options.last_block_offset;
         debug!("ctx": "aes", "op": "authenticate_text",
             "path": crypt_path, "memfd": memfd, "mode": file_mode,
-            "msg": "feed ciphertext into HMAC algoritm socket");
+            "msg": "feed ciphertext into HMAC algorithm socket");
 
         #[expect(clippy::cast_possible_wrap)]
         lseek64(&enc_fd, SYD3_HDR_SIZE as i64, Whence::SeekSet)?;
